/*
Copyright (C) 2014 Graham Jones, www.indriid.com
*/
package stacey.util;

import beast.core.Description;
import beast.evolution.tree.Node;
import beast.evolution.tree.Tree;
import beast.util.Randomizer;
import stacey.NodesNudge;

import java.util.*;

/**
 * Created by Graham Jones on 26/10/2014.
 */


/*
This is for operators to do calculations. It is sort of
parallel to PIOProbabilityGeneTree. It cooperates with OperatorsSMCTree.

A OperatorsGeneTree is created by the operator initially, then updated
as required. Each operator maintains a copy of each gene tree using one of these.

It has two inner classes: GeneUnionNode and GeneTreeSPRSpecification.
*/

@Description("A tree built from an underlying gene tree which is used by operators to choose details of moves")

public class OperatorsGeneTree {
    private final Tree tree;
    private final GeneUnionNode[] opGNodes;
    private final Bindings bindings;



    // data for moving one subtree
    private class GeneTreeSPRSpecification {
        public final Node source;
        public final double height;
        public final int choiceCount;
        public final Node destination;

        public GeneTreeSPRSpecification(Node source, double height, int choiceCount, Node destination) {
            this.source = source;
            this.height = height;
            this.choiceCount = choiceCount;
            this.destination = destination;
        }
    }


    // Node for a copy of node times and topology of a BEAST gene tree with unions at nodes
    public class GeneUnionNode implements UnionNode {
        private int lft;
        private int rgt;
        private int anc;
        private double height;
        private final BitUnion union;
        private Node beastgtreenode;
        private String ID; // for debugging

        // Constructor claims memory, makes empty unions.
        public GeneUnionNode(int smcTreeLeafCount) {
            lft = -1;
            rgt = -1;
            anc = -1;
            height = 0.0;
            union = new BitUnion(smcTreeLeafCount);
            beastgtreenode = null;
            ID = "";
        }


        public int getLft() { return lft; }
        public int getRgt() { return rgt; }
        public BitUnion getUnion() { return union; }
        public double getHeight() { return height; }
        public String getID() { return ID; }
        public String getExtraText() { return  (beastgtreenode==null ? "null" : "has beast node");  }


        public void setLft(int lft) { this.lft = lft; }
        public void setRgt(int rgt) { this.rgt = rgt; }
        public void setAnc(int anc) { this.anc = anc; }
        public void setHeight(double height) { this.height = height; }
        public void setBNode(Node bNode) { beastgtreenode = bNode; }

        public void setTipUnionAndID(String id) {
            this.ID = id;
            union.replaceWith(bindings.tipUnionFromAllTaxaID(id));
            /*int ati = bindings.allTaxaID2index(id);
            int smc = bindings.allTaxaIndex2smcIndex(ati);
            union.insert(smc);*/
        }


        // reverts node back to empty in preparation for updating from BEAST gene tree
        public void reset() {
            lft = -1;
            rgt = -1;
            anc = -1;
            height = 0.0;
            union.reset();
            beastgtreenode = null;
            ID = "";
        }

    }



    /****************************************************************/
    /********************  main class *********************************/
    /****************************************************************/


    // constructor claims memory then calls update
    public OperatorsGeneTree(Tree gtree, Bindings bindings) {
        this.tree = gtree;
        this.bindings = bindings;
        opGNodes = new GeneUnionNode[tree.getNodeCount()];
        int nSMCs = bindings.smcTreeTipCount();
        for (int i = 0; i < opGNodes.length; i++) {
            opGNodes[i] = new GeneUnionNode(nSMCs);
        }
        reset();
    }



    public void reset() {
        for (int i = 0; i < opGNodes.length; i++) {
            opGNodes[i].reset();
        }
    }


    // updates from BEAST gene tree
    public void update() {
        assert opGNodes[0].beastgtreenode == null;
        UnionNode.Abstract.beastTreeToUnionTree(opGNodes, tree);
    }




    /***************************  debug *********************/

    public OperatorsGeneTree.GeneUnionNode [] getNodes() {
        return opGNodes;
    }




    public String toString() {
        return UnionNode.Abstract.treeAsText(opGNodes, "beast_node");
    }



    /*****************************************************************************/
    /************************ Nodes nudge ****************************************/
    /*****************************************************************************/


    public ArrayList<Node> nodesToNudge(NudgeMoveSMCNodeInfo nani, NodesNudge.GtreeNodeCriterion criterion) {
        BitUnion spp = nani.getNodeUnion();
        BitUnion sppl = nani.getLftUnion();
        BitUnion sppr = nani.getRgtUnion();
        ArrayList<Node> nodes = new ArrayList<Node>(0);
        for (int g = 0; g < opGNodes.length; g++) {
            if (opGNodes[g].lft >= 0) {
                BitUnion gunion = opGNodes[g].union;
                if (gunion.isContainedIn(spp)) {
                    boolean mixed_with_pure_children = mixedWithPureChildren(opGNodes, g, sppl, sppr);
                    boolean mixed_with_a_pure_child = mixedWithAPureChild(opGNodes, g, sppl, sppr);
                    boolean mixed = mixedNode(opGNodes, g, sppl, sppr);
                    assert mixed || !mixed_with_pure_children;
                    assert mixed || !mixed_with_a_pure_child;
                    assert mixed_with_a_pure_child || !mixed_with_pure_children;

                    if (nodeWanted(criterion,
                            mixed_with_pure_children,
                            mixed_with_a_pure_child,
                            mixed)) {
                        nodes.add(opGNodes[g].beastgtreenode);
                    }
                }
            }
        }
        return nodes;
    }



    /***************************************************************************************/
    /***************************** FocusedNodeHeightScaler ********************************/
    /***************************************************************************************/


    public void fillInDistancesLogHeights(FSMoveSMCNodeUnions fsNUnions, int [] distances, double [] logHeights) {
        fillInDistances(fsNUnions, distances);
        fillInLogHeights(logHeights);
    }



    public void getFocusedScalerBounds(double[] interval, FSMoveSMCNodeInfo[] fsNodeInfos,
                                       double[] logHeights, double[] weights) {
        getFocusedScalerInternalBounds(interval, logHeights, weights);
        for (int i = 0; i < fsNodeInfos.length; i++) {
            getFocusedScalerCompatibilityBounds(
                    fsNodeInfos[i], interval, logHeights, weights);
        }
    }


    public double fsMoveDoScale(double logSF, double [] weights) {
        double logHR = 0.0;
        for (int s = 0; s < opGNodes.length; s++) {
            if (opGNodes[s].lft >= 0 && opGNodes[s].anc >= 0 && weights[s] > 0.0) {
                double logSFW = logSF * weights[s];
                opGNodes[s].beastgtreenode.setHeight(opGNodes[s].beastgtreenode.getHeight() * Math.exp(logSFW));
                logHR += logSFW;

            }
        }
        return logHR;
    }



    /**********************************************************************************/
    /**************************   CoordinatedPruneRegraft  ****************************/
    /**********************************************************************************/

    //  obtains a list of moves, sorts them, does them.
    public double doCoordinatedSPRMoves(OperatorsSMCTree.SPRmoveInfo sppSPRInfo) {
        ArrayList<GeneTreeSPRSpecification> gtsprss = makeListOfSPRSpecs(sppSPRInfo);
        Collections.sort(gtsprss, GTREESPRSPEC_ORDER);
        double logHR = 0;
        for (GeneTreeSPRSpecification gtspr : gtsprss) {
            doSPRmove(gtspr.source, gtspr.destination);
            logHR += Math.log(gtspr.choiceCount);
        }
        return  logHR;
    }


    //  obtains a list of moves, finds the HR, does not edit the tree. This is for the
    // reverse move to doCoordinatedSPRMoves().
    public double HRForCoordinatedSPRMoves(OperatorsSMCTree.SPRmoveInfo sppSPRInfo) {
        ArrayList<GeneTreeSPRSpecification> gtsprss = makeListOfSPRSpecs(sppSPRInfo);
        double logHR = 0;
        for (GeneTreeSPRSpecification gtspr : gtsprss) {
            logHR += Math.log(gtspr.choiceCount);
        }
        return  logHR;
    }



    // makes a list of moves (GeneTreeSPRSpecifications) for one gene tree.
    // called from OperatorsGeneTree.
    ArrayList<GeneTreeSPRSpecification> makeListOfSPRSpecs(OperatorsSMCTree.SPRmoveInfo sppSPRInfo) {
        ArrayList<GeneTreeSPRSpecification> gtsprss = new ArrayList<GeneTreeSPRSpecification>(0);
        for (int a = 0; a < opGNodes.length; a++) {
            double h = opGNodes[a].height;
            double hM = sppSPRInfo.heightsDancs[sppSPRInfo.heightsDancs.length-1];
            if (opGNodes[a].lft >= 0  &&  h <= hM) {
                int x = opGNodes[a].lft;
                int y = opGNodes[a].rgt;
                boolean xinS = opGNodes[x].union.isContainedIn(sppSPRInfo.sppS);
                boolean yinS = opGNodes[y].union.isContainedIn(sppSPRInfo.sppS);
                if (xinS && !yinS) {
                    GeneTreeSPRSpecification xspr = chooseDestinationAndConstructMove(
                            a, x, sppSPRInfo);
                    gtsprss.add(xspr);
                }
                if (yinS && !xinS) {
                    GeneTreeSPRSpecification yspr = chooseDestinationAndConstructMove(
                            a, y, sppSPRInfo);
                    gtsprss.add(yspr);
                }
            }
        }
        return gtsprss;
    }



    /**********************************************************************************/
    /**************************  private  *********************************************/
    /**********************************************************************************/

    /**************************************************************************************/
    /********************* private for NodesNudge *****************************************/



    // next 3 methods test various topological properties of nodes near smc tree node
    // which has tips in spp, and whose children have tips in sppl, sppr

    // this one is first meetings
    private boolean mixedWithPureChildren(GeneUnionNode[] gunodes, int g, BitUnion sppl, BitUnion sppr) {
        int gl = gunodes[g].lft;
        int gr = gunodes[g].rgt;
        boolean linl = gunodes[gl].union.isContainedIn(sppl);
        boolean rinl = gunodes[gr].union.isContainedIn(sppl);
        boolean linr = gunodes[gl].union.isContainedIn(sppr);
        boolean rinr = gunodes[gr].union.isContainedIn(sppr);
        return ((linl && rinr) || (linr && rinl));
    }


    // this is a meeting where one child may be mixed but not both
    private boolean mixedWithAPureChild(GeneUnionNode[] gunodes, int g, BitUnion sppl, BitUnion sppr) {
        BitUnion gunion = gunodes[g].union;
        if (gunion.overlaps(sppl) && gunion.overlaps(sppr)) {
            int gl = gunodes[g].lft;
            int gr = gunodes[g].rgt;
            boolean linl = gunodes[gl].union.isContainedIn(sppl);
            boolean rinl = gunodes[gr].union.isContainedIn(sppl);
            boolean linr = gunodes[gl].union.isContainedIn(sppr);
            boolean rinr = gunodes[gr].union.isContainedIn(sppr);
            return (linl || rinr || linr || rinl);
        }
        return false;
    }


    // any meeting
    private boolean mixedNode(GeneUnionNode[] gunodes, int g, BitUnion sppl, BitUnion sppr) {
        BitUnion gunion = gunodes[g].union;
        return (gunion.overlaps(sppl) && gunion.overlaps(sppr));
    }


    // various combinations of properties defined in last 5 methods
    private boolean nodeWanted(NodesNudge.GtreeNodeCriterion criterion,
                               boolean mixed_with_pure_children,
                               boolean mixed_with_a_pure_child,
                               boolean mixed) {
        boolean wanted;
        switch (criterion) {
            case MIXED_WITH_PURE_CHILDREN:
                wanted = mixed_with_pure_children;
                break;

            case MIXED_WITH_A_PURE_CHILD:
                wanted = mixed_with_a_pure_child;
                break;

            case MIXED:
                wanted = mixed;
                break;

            default:
                wanted = false;
                assert false;
        }
        return wanted;
    }






    /********************* private FocusedNodeHeightScaler ****************************/


    private void fillInDistances(FSMoveSMCNodeUnions fsNUnions, int [] distances) {
        assert opGNodes.length == distances.length;
        BitUnion spp = fsNUnions.getNodeUnion();
        BitUnion sppl = fsNUnions.getLftUnion();
        BitUnion sppr = fsNUnions.getRgtUnion();
        for (int g = 0; g < opGNodes.length; g++) {
            if (gNodeSNodeCloselyLinked(g, spp, sppl, sppr)) {
                distances[g] = 1;
            } else if (gNodeSNodeLinked(g, sppl, sppr)) {
                distances[g] = 2;
            } else {
                distances[g] = Integer.MAX_VALUE;
            }
        }
        for (int g = 0; g < opGNodes.length; g++) {
            if (distances[g] < Integer.MAX_VALUE) {
                while (opGNodes[g].anc >= 0) {
                    int anc = opGNodes[g].anc;
                    distances[anc] = Math.min(distances[anc], distances[g] + 1);
                    g = anc;
                }
            }

        }
        fillInSubtreeDistances(opGNodes.length - 1, distances);
    }



    private void fillInLogHeights(double [] logHeights) {
        assert opGNodes.length == logHeights.length;
        for (int i = 0; i < opGNodes.length; i++) {
            if (opGNodes[i].lft >= 0) {
                logHeights[i] = Math.log(opGNodes[i].height);
            }
        }

    }


    private void fillInSubtreeDistances(int n, int [] distances) {
        if (opGNodes[n].anc >= 0) {
            int anc = opGNodes[n].anc;
            distances[n] = Math.min(distances[n], distances[anc] + 1);
        }
        if (opGNodes[n].lft >= 0) {
            fillInSubtreeDistances(opGNodes[n].lft, distances);
            fillInSubtreeDistances(opGNodes[n].rgt, distances);
        }
    }


    private void getFocusedScalerInternalBounds(double[] interval, double[] logHeights, double[] weights) {
        for (int n = 0; n < opGNodes.length; n++) {
            if (opGNodes[n].lft >= 0  &&  opGNodes[n].anc >= 0) {
                int anc = opGNodes[n].anc;
                double weightDiff = weights[n] - weights[anc];
                double loghDiff = logHeights[anc] - logHeights[n];
                if (weightDiff > 0) {
                    interval[1] = Math.min(interval[1], loghDiff/weightDiff);
                }
                if (weightDiff < 0) {
                    interval[0] = Math.max(interval[0], loghDiff/weightDiff);
                }
            }
        }
    }



    private void getFocusedScalerCompatibilityBounds(
            FSMoveSMCNodeInfo fsNodeInfo, double[] interval,
            double[] logHeights, double[] weights) {
        BitUnion sppL = fsNodeInfo.getLftUnion();
        BitUnion sppR = fsNodeInfo.getRgtUnion();
        double sLogHeight = fsNodeInfo.getLogHeight();
        double sWeight = fsNodeInfo.getWeight();
        for (int g = 0; g < opGNodes.length; g++) {
            if (gNodeSNodeLinked(g, sppL, sppR)) {
                double weightDiff = sWeight - weights[g];
                double loghDiff = logHeights[g] - sLogHeight;
                if (weightDiff > 0) {
                    interval[1] = Math.min(interval[1], loghDiff/weightDiff);
                }
                if (weightDiff < 0) {
                    interval[0] = Math.max(interval[0], loghDiff/weightDiff);
                }
            }
        }
    }



    private boolean gNodeSNodeCloselyLinked(int g, BitUnion spp, BitUnion sppL, BitUnion sppR) {
        if (opGNodes[g].lft >= 0  &&  opGNodes[g].union.isContainedIn(spp)) {
            int lft = opGNodes[g].lft;
            int rgt =  opGNodes[g].rgt;
            boolean LwithinL = opGNodes[lft].union.isContainedIn(sppL);
            boolean RwithinL = opGNodes[rgt].union.isContainedIn(sppL);
            boolean LwithinR = opGNodes[lft].union.isContainedIn(sppR);
            boolean RwithinR = opGNodes[rgt].union.isContainedIn(sppR);
            return ((LwithinL & RwithinR) | (LwithinR && RwithinL));
        } else {
            return false;
        }
    }




    private boolean gNodeSNodeLinked(int g, BitUnion sppL, BitUnion sppR) {
        if (opGNodes[g].lft >= 0) {
            boolean overlapsL = opGNodes[g].union.overlaps(sppL);
            boolean overlapsR = opGNodes[g].union.overlaps(sppR);
            boolean linked = overlapsL & overlapsR;
            int lft = opGNodes[g].lft;
            int rgt =  opGNodes[g].rgt;
            boolean LoverlapsL = opGNodes[lft].union.overlaps(sppL);
            boolean LoverlapsR = opGNodes[lft].union.overlaps(sppR);
            boolean Llinked = LoverlapsL & LoverlapsR;
            boolean RoverlapsL = opGNodes[rgt].union.overlaps(sppL);
            boolean RoverlapsR = opGNodes[rgt].union.overlaps(sppR);
            boolean Rlinked = RoverlapsL & RoverlapsR;
            return (linked & !Llinked & !Rlinked);
        } else {
            return false;
        }
    }



    /********************* private CoordinatedPruneRegraft  ****************************/





    // makes a list of possible destinations for one gene subtree
    // then chooses one and returns a GeneTreeSPRSpecification
    // (which is source, destination, height, number of choices).
    private GeneTreeSPRSpecification chooseDestinationAndConstructMove(
            int anc, int s,
            OperatorsSMCTree.SPRmoveInfo sppSPRInfo) {

        Node source = opGNodes[s].beastgtreenode;
        double height = opGNodes[anc].height;

        int nDs = sppSPRInfo.heightsDancs.length;
        assert height <= sppSPRInfo.heightsDancs[nDs-1];
        int d = 0;
        while (height > sppSPRInfo.heightsDancs[d]) {
            d++;
        }
        ArrayList<Node> possibleDestinations = new ArrayList<Node>();
        for (int x = 0; x < opGNodes.length; x++) {
            if (opGNodes[x].height <= height  &&  height <= opGNodes[ opGNodes[x].anc ].height  &&
                    opGNodes[x].union.isContainedIn(sppSPRInfo.sppDs[d])) {
                possibleDestinations.add(opGNodes[x].beastgtreenode);
            }
        }
        int choiceCount = possibleDestinations.size();
        assert choiceCount > 0;
        Node destination = possibleDestinations.get(Randomizer.nextInt(choiceCount));
        GeneTreeSPRSpecification spr = new GeneTreeSPRSpecification(source, height, choiceCount, destination);
        return spr;
    }




    //  this one actually edits the BEAST gene tree
    private void doSPRmove(Node s, Node d) {
        Node a = s.getParent();
        Node x = a.getChild(0) == s ? a.getChild(1) : a.getChild(0);
        Node y = a.getParent();
        Node z = d.getParent();
        z.removeChild(d);
        a.removeChild(x);
        y.removeChild(a);
        z.addChild(a);
        a.addChild(d);
        y.addChild(x);
    }



    // used for sorting the SPR moves from most ancient to most recent.
    private static final Comparator<OperatorsGeneTree.GeneTreeSPRSpecification> GTREESPRSPEC_ORDER =
            new Comparator<OperatorsGeneTree.GeneTreeSPRSpecification>() {
                public int compare(OperatorsGeneTree.GeneTreeSPRSpecification a, OperatorsGeneTree.GeneTreeSPRSpecification b) {
                    if (a.height != b.height) {
                        if (a.height < b.height) {
                            return 1;
                        } else {
                            return -1;
                        }
                    } else {
                        return 0;
                    }
                }
            };




}



/****************************** Some NodesNudge experiments ****************************************/

/*

    public ArrayList<Node> nodesToNudge(NudgeMoveSMCNodeInfo nani, NodesNudge.GtreeNodeCriterion criterion) {
        BitUnion spp = nani.getNodeUnion();
        BitUnion sppl = nani.getLftUnion();
        BitUnion sppr = nani.getRgtUnion();
        ArrayList<Node> nodes = new ArrayList<Node>(0);
        for (int g = 0; g < opGNodes.length; g++) {
            if (opGNodes[g].lft >= 0) {
                BitUnion gunion = opGNodes[g].union;
                if (gunion.isContainedIn(spp)) {
                    boolean mixed_with_pure_children = mixedWithPureChildren(opGNodes, g, sppl, sppr);
                    boolean mixed_with_a_pure_child = mixedWithAPureChild(opGNodes, g, sppl, sppr);
                    boolean mixed = mixedNode(opGNodes, g, sppl, sppr);
                    boolean pure_node_mixed_parent = pureNodeMixedParent(opGNodes, g, sppl, sppr);
                    boolean parent_is_in_group = parentIsIngroup(opGNodes, g, spp);
                    assert mixed || !mixed_with_pure_children;
                    assert mixed || !mixed_with_a_pure_child;
                    assert mixed_with_a_pure_child || !mixed_with_pure_children;

                    if (nodeWanted(criterion,
                            mixed_with_pure_children,
                            mixed_with_a_pure_child,
                            mixed, pure_node_mixed_parent,
                            parent_is_in_group)) {
                        nodes.add(opGNodes[g].beastgtreenode);
                    }
                }
            }
        }
        return nodes;
    }
*/





/*
    // the node has not met 'other side' but parent has
    private boolean pureNodeMixedParent(GeneUnionNode[] gunodes, int g, BitUnion sppl, BitUnion sppr) {
        BitUnion gunion = gunodes[g].union;
        boolean want = false;
        if (gunion.isContainedIn(sppl) || gunion.isContainedIn(sppr)) {
            int ga = gunodes[g].anc;
            assert ga >= 0;
            BitUnion gaunion = gunodes[ga].union;
            if (gaunion.overlaps(sppl) && gaunion.overlaps(sppr)) {
                want = true;
            }
        }
        return want;
    }


    // parent is in spp, ie has not met tips outside clade at smc node
    private boolean parentIsIngroup(GeneUnionNode[] gunodes, int g, BitUnion spp) {
        int ga = gunodes[g].anc;
        if (ga < 0) {
            return true;
        } else {
            BitUnion gaunion = gunodes[ga].union;
            return gaunion.isContainedIn(spp);
        }

    }*/







/*

    // various combinations of properties defined in last 5 methods
    private boolean nodeWanted(NodesNudge.GtreeNodeCriterion criterion,
                               boolean mixed_with_pure_children,
                               boolean mixed_with_a_pure_child,
                               boolean mixed,
                               boolean pure_node_mixed_parent,
                               boolean parent_is_in_group) {
        boolean wanted = false;
        switch (criterion) {
            case MIXED_WITH_PURE_CHILDREN_AND_PARENT_IN_GROUP:
                if (mixed_with_pure_children && parent_is_in_group) {
                    wanted = true;
                }

            case MIXED_WITH_PURE_CHILDREN:
                if (mixed_with_pure_children) {
                    wanted = true;
                }
                break;

            case MIXED_WITH_A_PURE_CHILD:
                if (mixed_with_a_pure_child) {
                    wanted = true;
                }
                break;

            case MIXED:
                if (mixed) {
                    wanted = true;
                }
                break;

            case PURE_MIXED_PARENT:
                if (pure_node_mixed_parent) {
                    wanted = true;
                }
                break;

            case PURE_MIXED_PARENT_OR_MIXED_WITH_PURE_CHILDREN:
                if (pure_node_mixed_parent || mixed_with_pure_children) {
                    wanted = true;
                }
                break;

            case PURE_MIXED_PARENT_OR_MIXED_WITH_A_PURE_CHILD:
                if (pure_node_mixed_parent || mixed_with_a_pure_child) {
                    wanted = true;
                }
                break;

            case PURE_MIXED_PARENT_OR_MIXED:
                if (pure_node_mixed_parent || mixed) {
                    wanted = true;
                }
                break;

            default:
                assert false;
        }
        return wanted;
    }
*/














